using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;
using Utils.Logs.Event;
using System.ComponentModel;
using System.IO;
using Utils.Logs.Exceptions;
using System.Timers;
using Utils.Misc;

namespace Utils.Logs
{
    /// <summary>
    /// Contain a list of log entry
    /// </summary>
    public class LogSet : IEnumerator, IDisposable, IListSource
    {
        #region Variables
        /// <summary>
        /// Interval of the timer
        /// </summary>
        private const int TIMER_INTERVAL = 1000;
        /// <summary>
        /// Timer to check whether the file changed
        /// </summary>
        private Timer vTimer = new Timer(TIMER_INTERVAL);
        /// <summary>
        /// List of log entries
        /// </summary>
        private ArrayList vList = new ArrayList();
        /// <summary>
        /// Path if the log file
        /// </summary>
        private string vFilePath = "";
        /// <summary>
        /// IEnumarator index
        /// </summary>
        private int vIndex = -1;
        /// <summary>
        /// Indicates whether the object is analyzing the log file
        /// </summary>
        private bool vGuarding = false;
        /// <summary>
        /// Indicates the last modification time of the file
        /// </summary>
        private DateTime vLastFileModification;
        /// <summary>
        /// Indicates the size of the file
        /// </summary>
        private float vFileSize = 0;
        /// <summary>
        /// Indicates the indexes with the bigges row (Date, Type, Title, Message)
        /// </summary>
        private ColumnsSize vColumnsSize;
        /// <summary>
        /// Selected LogTypes in the Log
        /// </summary>
        private LogType[] vLogTypes;
        #endregion

        #region Constructor
        /// <summary>
        /// Constructor
        /// </summary>
        public LogSet()
        {
            vColumnsSize.SetDefaultValues();
            SetAllLogType();
        }
        /// <summary>
        /// Constructor
        /// </summary>
        /// <param name="_FilePath">Indicates the path of the Logs file</param>
        public LogSet(string _FilePath)
        {
            this.FilePath = _FilePath;
            vColumnsSize.SetDefaultValues();
            SetAllLogType();
        }
        /// <summary>
        /// Dispose the ressouces
        /// </summary>
        ~LogSet() { this.Dispose(); }
        #endregion

        #region Properties
        /// <summary>
        /// Gets or sets the list of log entries
        /// </summary>
        public ArrayList List
        {
            get { return vList; }
            set { vList = value; }
        }
        /// <summary>
        /// Gets the elements number _index from the log entry list
        /// </summary>
        /// <param name="_index">Number of the log entry</param>
        /// <returns>The log entry</returns>
        /// <exception cref="IndexOutOfRangeException">If the index if too big or too small</exception>
        public LogRow this[int _index]
        {
            get
            {
                if (_index > vList.Count || _index < 0)
                    throw new IndexOutOfRangeException("The index does not exist in the log entry list");
                else
                    return (LogRow)vList[_index];

            }
        }
        /// <summary>
        /// Gets the number of log entries in the list
        /// </summary>
        public int Count { get { return vList.Count; } }
        /// <summary>
        /// Gets or sets the path of the log file
        /// </summary>
        public string FilePath
        {
            get { return vFilePath; }
            set { vFilePath = value; }
        }
        /// <summary>
        /// Gets the last modification of the log file
        /// </summary>
        public DateTime LastFileModification
        {
            get { return vLastFileModification; }
            private set { vLastFileModification = value; }
        }
        /// <summary>
        /// Gets the dates contained in the logs
        /// </summary>
        public DateTime[] ListOfDates
        {
            get
            {
                List<DateTime> Dates = new List<DateTime>();
                foreach (LogRow lr in this.List)
                {
                    DateTime lrDate = new DateTime(lr.Date.Year, lr.Date.Month, lr.Date.Day);
                    if (!Dates.Contains(lrDate))
                        Dates.Add(lrDate);
                }
                DateTime[] dt = new DateTime[Dates.Count];
                for (int i = 0; i < dt.Length; i++)
                {
                    dt[i] = Dates[i];
                }
                return dt;
            }
        }
        /// <summary>
        /// Gets the list of LogTypes selected in the LogSet
        /// </summary>
        public LogType[] ListOfTypes { get { return this.vLogTypes; } }
        /// <summary>
        /// Gets the log file size in Ko
        /// </summary>
        public float FileSize { get { return this.vFileSize / 1024; } }
        /// <summary>
        /// Gets the index and the size of the biggest columns of the logs
        /// </summary>
        public ColumnsSize ColumnsSize { get { return vColumnsSize; } }
        #endregion

        #region Methods
        /// <summary>
        /// Fill the list of log entries from the file path
        /// </summary>
        /// <exception cref="LogException">The file does not contain expected log entries</exception>
        public void Fill()
        {
            int vID = 0;
            LogRow lr = new LogRow();
            string vBuffer = ReadFile();
            vBuffer = vBuffer.Replace(Log.WLogTag, "");
            string[] opt = { Log.LineSeparator };
            string[] vEntryList = vBuffer.Split(opt, StringSplitOptions.RemoveEmptyEntries);
            this.List.Clear();
            //foreach (string vLine in vEntryList)
            for (int i = 0; i < vEntryList.Length; i++)
            {
                try
                {
                    lr = LogRow.Parse(vEntryList[i]);
                    lr.ID = vID;
                    vID++;
                    this.Add(lr);

                    #region Biggest
                    int DateLength = lr.Date.ToString("dd/MM/yyyy [HH:mm:ss]").Length;
                    int TypeLength = lr.Type.ToString().Length;
                    int TitleLength = lr.Title.Length;
                    int MessageLenth = lr.Message.Length;

                    if (DateLength > vColumnsSize.DateLength)
                    {
                        vColumnsSize.DateLength = DateLength;
                        vColumnsSize.DateIndex = i;
                        vColumnsSize.DateValue = lr.Date.ToString("dd/MM/yyyy [HH:mm:ss]");
                    }
                    if (TypeLength > vColumnsSize.TypeLength)
                    {
                        vColumnsSize.TypeLength = TypeLength;
                        vColumnsSize.TypeIndex = i;
                        vColumnsSize.TypeValue = lr.Type.ToString();
                    }
                    if (TitleLength > vColumnsSize.TitleLength)
                    {
                        vColumnsSize.TitleLength = TitleLength;
                        vColumnsSize.TitleIndex = i;
                        vColumnsSize.TitleValue = lr.Title;
                    }
                    if (MessageLenth > vColumnsSize.MessageLenth)
                    {
                        vColumnsSize.MessageLenth = MessageLenth;
                        vColumnsSize.MessageIndex = i;
                        vColumnsSize.MessageValue = lr.Message;
                    }
                    #endregion
                }
                catch (BadLogParseException ex)
                { throw new LogException(ex.Message); }
            }
            SetAllLogType();
        }
        /// <summary>
        /// Keeps only the row with the LogType given as argument
        /// </summary>
        /// <param name="_LogTypes">LogType to keep in the LogSet</param>
        public void KeepOnly(LogType[] _LogTypes)
        {
            this.Fill();
            ArrayList Buffer = new ArrayList();
            foreach (LogRow item in this.vList)
            {
                foreach (LogType l in _LogTypes)
                {
                    if (item.Type == l)
                    {
                        Buffer.Add(item);
                        break;
                    }
                }
            }
            this.vList = Buffer;
            this.vLogTypes = _LogTypes;
        }
        /// <summary>
        /// Add a log entry in the list
        /// </summary>
        /// <param name="_LogRow">Logs entry to add in the list</param>       
        public void Add(LogRow _LogRow) { this.vList.Add(_LogRow); }
        /// <summary>
        /// Add a log entry list in the list
        /// </summary>
        /// <param name="_LogSet">Logs entry list to add in the list</param>
        public void AddRange(LogSet _LogSet)
        {
            //this.vList. AddRange(_LogSet.vList); 
            foreach (LogRow lr in _LogSet)
            {
                this.Add(lr);
            }
        }
        /// <summary>
        /// Delete the log entry at the index _index
        /// </summary>
        /// <param name="_index">_index of the log entry to delete</param>        
        /// <exception cref="ArgumentOutOfRangeException">Impossible to remove the entry at the given position</exception>
        public void RemoveAt(int _index)
        {
            try { this.List.RemoveAt(_index); }
            catch (ArgumentOutOfRangeException)
            {
                throw new ArgumentOutOfRangeException("Impossible to remove the entry at position " + _index);
            }
        }
        /// <summary>
        /// Save the list into a log file
        /// </summary>
        /// <exception cref="IOException">The exception that is thrown when an I/O error occurs.</exception>
        /// <exception cref="Exception">An unexpected error occured</exception>
        public void Save()
        {
            string vToWrite = Log.WLogTag + Log.LineSeparator;

            foreach (LogRow lr in this.List)
                vToWrite += lr;
            try
            {
                FileInfo file = new FileInfo(this.FilePath);
                FileStream fs = file.Create();
                StreamWriter sw = new StreamWriter(fs);
                sw.Write(vToWrite);
                sw.Close();
                fs.Close();
            }
            catch (IOException) { throw; }
            catch (Exception) { throw; }
        }
        /// <summary>
        /// Reset the whole list of log entries
        /// </summary>
        public void Clear()
        {
            vList = new ArrayList();
        }
        /// <summary>
        /// Search all the log of this date
        /// </summary>
        /// <param name="_Time">Date of the logs</param>
        /// <returns>LogSet which contains the results of the search</returns>
        public LogSet Search(DateTime _Time)
        {
            LogSet _Result = new LogSet(this.FilePath);
            foreach (LogRow lr in this.List)
            {
                if (lr.Date.Year == _Time.Year
                    && lr.Date.Month == _Time.Month
                    && lr.Date.Day == _Time.Day)
                    _Result.Add(lr);
            }
            return _Result;
        }
        /// <summary>
        /// Search all the log of this type
        /// </summary>
        /// <param name="_Type">Type of log </param>
        /// <returns>LogSet which contains the results of the search</returns>
        public LogSet Search(LogType _Type)
        {
            LogSet _Result = new LogSet(this.FilePath);
            foreach (LogRow lr in this.List)
            {
                if (lr.Type == _Type)
                    _Result.Add(lr);
            }
            return _Result;
        }
        /// <summary>
        /// Search all the log of these types
        /// </summary>
        /// <param name="_Types">Array of log types</param>
        /// <returns>LogSet which contains the results of the search</returns>
        public LogSet Search(List<LogType> _Types)
        {
            LogSet _Result = new LogSet(this.FilePath);
            foreach (LogType t in _Types)
            {
                LogSet ls = this.Search(t);
                _Result.AddRange(ls);
            }
            return _Result;
        }
        /// <summary>
        /// Analyze the log file and throws the event FileChanged if the file changed
        /// </summary>
        /// <exception cref="ArgumentException">The file name is empty, contains only white spaces, 
        /// or contains invalid characters.</exception>
        /// <exception cref="NotSupportedException">fileName contains a colon (:) in the middle of the string.</exception>
        /// <exception cref="PathTooLongException">The specified path, file name, or both exceed the system-defined 
        /// maximum length. For example, on Windows-based platforms, paths must be less than 248 characters, and 
        /// file names must be less than 260 characters.</exception>
        /// <exception cref="UnauthorizedAccessException">Access to fileName is denied.</exception>
        /// <exception cref="Exception">Unexpected error</exception>
        public void StartGuard()
        {
            try
            {
                if (!vGuarding)
                {
                    SetupTimer();
                    vTimer.Start();
                    vGuarding = true;
                }
            }
            catch (ArgumentException) { throw; }
            catch (NotSupportedException) { throw; }
            catch (PathTooLongException) { throw; }
            catch (UnauthorizedAccessException) { throw; }
            catch (Exception) { throw; }
        }
        /// <summary>
        /// Stop the analyze the log file.
        /// </summary>
        public void StopGuard()
        {
            vTimer.Stop();
            vGuarding = false;
        }
        /// <summary>
        /// Give the next element of the list or throw an exception if does not exist
        /// </summary>
        /// <exception cref="ArgumentOutOfRangeException">If the next element does not exist</exception>
        public LogRow Next()
        {
            try { return (LogRow)this.List[++vIndex]; }
            catch (ArgumentOutOfRangeException) { throw; }
        }
        /// <summary>
        /// Give the previous element of the list or throw an exception if does not exist
        /// </summary>
        /// <exception cref="ArgumentOutOfRangeException">If the previous element does not exist</exception>
        public LogRow Previous()
        {
            try { return (LogRow)this.List[--vIndex]; }
            catch (ArgumentOutOfRangeException) { throw; }
        }
        /// <summary>
        /// Give the next element of the list or NULL if does not exist
        /// </summary>
        public LogRow TryNext()
        {
            try { return this.Next(); }
            catch (ArgumentOutOfRangeException) { return null; }
        }
        /// <summary>
        /// Give the previous element of the list or NULL if does not exist
        /// </summary>
        ///         
        public LogRow TryPrevious()
        {
            try { return this.Previous(); }
            catch (ArgumentOutOfRangeException) { return null; }
        }
        /// <summary>
        ///Set all the LogType int the vLogType list
        /// </summary>
        private void SetAllLogType()
        {
            this.vLogTypes = new LogType[Enum.GetValues(typeof(LogType)).Length];
            int i = 0;
            foreach (LogType item in Enum.GetValues(typeof(LogType)))
                vLogTypes[i++] = item;
        }
        #region Private
        /// <summary>
        /// Read the file and returns the text
        /// </summary>
        /// <returns>The text of the file</returns>
        /// <exception cref="ArgumentException">The file name is empty, contains only white spaces, 
        /// or contains invalid characters.</exception>
        /// <exception cref="NotSupportedException">fileName contains a colon (:) in the middle of the string.</exception>
        /// <exception cref="PathTooLongException">The specified path, file name, or both exceed the system-defined 
        /// maximum length. For example, on Windows-based platforms, paths must be less than 248 characters, and 
        /// file names must be less than 260 characters.</exception>
        /// <exception cref="UnauthorizedAccessException">Access to fileName is denied.</exception>
        /// <exception cref="Exception">Unexpected error</exception>
        private string ReadFile()
        {
            try
            {
                FileInfo fi = new FileInfo(this.FilePath);
                this.vFileSize = fi.Length;
                if (!fi.Exists)
                    throw new FileNotFoundException("The log file does not exists at " + this.FilePath);
                StreamReader sr = fi.OpenText();
                string ret = sr.ReadToEnd();
                sr.Close();
                return ret;
            }
            catch (ArgumentException) { throw; }
            catch (NotSupportedException) { throw; }
            catch (PathTooLongException) { throw; }
            catch (UnauthorizedAccessException) { throw; }
            catch (Exception) { throw; }
        }
        /// <summary>
        /// Setup the BackgroundWorker
        /// </summary>
        /// <exception cref="ArgumentException">The file name is empty, contains only white spaces, 
        /// or contains invalid characters.</exception>
        /// <exception cref="NotSupportedException">fileName contains a colon (:) in the middle of the string.</exception>
        /// <exception cref="PathTooLongException">The specified path, file name, or both exceed the system-defined 
        /// maximum length. For example, on Windows-based platforms, paths must be less than 248 characters, and 
        /// file names must be less than 260 characters.</exception>
        /// <exception cref="UnauthorizedAccessException">Access to fileName is denied.</exception>
        /// <exception cref="Exception">Unexpected error</exception>
        private void SetupTimer()
        {
            try
            {
                lock (this)
                {
                    this.LastFileModification = File.GetLastWriteTime(this.FilePath);
                    vTimer.Elapsed += new ElapsedEventHandler(Event_TimerElapsed);
                }
            }
            catch (ArgumentException) { throw; }
            catch (NotSupportedException) { throw; }
            catch (PathTooLongException) { throw; }
            catch (UnauthorizedAccessException) { throw; }
            catch (Exception) { throw; }
        }
        #endregion
        #endregion

        #region Events
        /// <summary>
        /// Thrown if the file changed. Check if the last modification time changed
        /// </summary>
        public event FileChangedEventHandler FileChanged;
        /// <summary>
        /// Notify a file changed
        /// </summary>
        /// <param name="args"></param>
        protected void OnFileChanged(FileChangedEventArgs args)
        {
            if (FileChanged != null)
                FileChanged(this, args);
        }
        /// <summary>
        /// Occures when the timer ticks
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Event_TimerElapsed(object sender, ElapsedEventArgs e)
        {
            lock (this)
            {
                DateTime dtBuffer = File.GetLastWriteTime(this.FilePath);
                if (dtBuffer > this.LastFileModification)
                {
                    this.LastFileModification = dtBuffer;
                    FileChangedEventArgs a = new FileChangedEventArgs();
                    a.DateTimeOfChange = dtBuffer;
                    a.DateTimeOfEvent = DateTime.Now;
                    OnFileChanged(a);
                }
            }
        }
        #endregion

        #region IEnumerator Members
        /// <summary>
        /// Return the current element
        /// </summary>
        public object Current { get { return vList[vIndex]; } }
        /// <summary>
        /// Move to the next element
        /// </summary>
        /// <returns></returns>
        public bool MoveNext() { return ++vIndex < vList.Count; }
        /// <summary>
        /// Reset the index of the enumaration
        /// </summary>
        public void Reset() { vIndex = -1; }
        /// <summary>
        /// Get the enumerator
        /// </summary>
        /// <returns></returns>
        public IEnumerator GetEnumerator() { return this; }
        #endregion

        #region IDisposable Members
        /// <summary>
        /// Dispose the ressources
        /// </summary>
        public void Dispose()
        {
            this.vTimer.Stop();
            this.vTimer.Dispose();
            this.FileChanged = null;
        }
        #endregion

        #region IListSource Members
        /// <summary>
        /// Gets a value indicating whether the collection is a collection of System.Collections.IList
        /// objects.
        /// </summary>
        public bool ContainsListCollection { get { return true; } }// (vList != null); } }
        /// <summary>
        /// Returns an System.Collections.IList that can be bound to a data source from
        /// an object that does not implement an System.Collections.IList itself.
        /// </summary>
        /// <returns>An System.Collections.IList that can be bound to a data source from the object.</returns>
        public IList GetList() { return this.List; }
        #endregion
    }
    /// <summary>
    /// Indicates the size of the biggest columns and their indexes
    /// </summary>
    public struct ColumnsSize
    {
        /// <summary>
        /// Index of the biggest date
        /// </summary>
        public int DateIndex;
        /// <summary>
        /// Length of the biggest date
        /// </summary>
        public int DateLength;
        /// <summary>
        /// Value of the biggest date
        /// </summary>
        public string DateValue;
        /// <summary>
        /// Index of the biggest type
        /// </summary>
        public int TypeIndex;
        /// <summary>
        /// Length of the biggest type
        /// </summary>
        public int TypeLength;
        /// <summary>
        /// Value of the biggest type
        /// </summary>
        public string TypeValue;
        /// <summary>
        /// Index of the biggest title
        /// </summary>
        public int TitleIndex;
        /// <summary>
        /// Length of the biggest title
        /// </summary>
        public int TitleLength;
        /// <summary>
        /// Value of the biggest title
        /// </summary>
        public string TitleValue;
        /// <summary>
        /// Index of the biggest message
        /// </summary>
        public int MessageIndex;
        /// <summary>
        /// Length of the biggest message
        /// </summary>
        public int MessageLenth;
        /// <summary>
        /// Value of the biggest message
        /// </summary>
        public string MessageValue;
        /// <summary>
        /// Sets the default values of thr structure
        /// </summary>
        public void SetDefaultValues()
        {
            DateIndex = -1;
            DateLength = 0;
            DateValue = "";

            TypeIndex = -1;
            TypeLength = 0;
            TypeValue = "";

            TitleIndex = -1;
            TitleLength = 0;
            TitleValue = "";

            MessageIndex = -1;
            MessageLenth = 0;
            MessageValue = "";
        }
    }
}
